/*
 * Copyright (c) 2014-2016 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Fri May 23 2014
 */
#include "ac_tk3_flash.h"

#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "flash.h"

static int			mkbl_connect(int fd);
static const struct avr_dev *	mkbl_device(int fd);
static int			mkbl_write(int fd, const struct avr_dev *dev,
                                           char *data, size_t size,
                                           enum memk kind);
static void *			mkbl_eeprom(int fd, const struct avr_dev *dev);
static int			mkbl_exit(int fd);

static const struct avr_dev {
  const char *name;
  unsigned char id;
  size_t flash_size, eeprom_size;
} avr_devs[] = {
  { .name = "ATMEGA1284P", .id = 0x7a,
    .flash_size = 0x20000, .eeprom_size = 0x1000 },
  { .name = "ATMEGA168", .id = 0x78,
    .flash_size = 0x4000, .eeprom_size = 0x200 },
  { NULL, 0, 0 }
};


/* --- mkbl_flash ---------------------------------------------------------- */

int
mkbl_flash(const char *exe, const char *port, int baud, int init_p)
{
  const struct avr_dev *dev;
  size_t size;
  char *data;
  Elf *elf;
  int fd;

  elf = elf_init(exe);
  if (!elf) return -1;

  fd = open_tty(port, baud);
  if (fd < 0) { warn("%s", port); return -1; }

  if (mkbl_connect(fd)) return -1;
  dev = mkbl_device(fd);
  if (!dev) return -1;

  data = elf_loadable_data(elf, &size, FL_FLASH);
  if (!data) return -1;
  if (mkbl_write(fd, dev, data, size, FL_FLASH)) return -1;
  free(data);

  if (init_p) {
    data = elf_loadable_data(elf, &size, FL_EEPROM);
    if (!data) return -1;
    if (!size) {
      data = strdup(TK3_MAGIC);
      size = sizeof(TK3_MAGIC);
    }
    if (mkbl_write(fd, dev, data, size, FL_EEPROM)) return -1;
    free(data);
  }

  /* display eeprom data */
  data = mkbl_eeprom(fd, dev);
  if (!data) return -1;

  if (strcmp(data, TK3_MAGIC)) {
    warnx("invalid eeprom data, please flash with option -i");
    return -1;
  }
  print_settings(data + sizeof(TK3_MAGIC));
  free(data);

  if (mkbl_exit(fd)) return -1;
  return 0;
}


/* --- mkbl_params --------------------------------------------------------- */

int
mkbl_params(const char *port, int baud, char **argv)
{
  const struct avr_dev *dev;
  struct {
    char key[128];
    int32_t value;
    char descr[128];
  } settings[128];
  char *data, *p, *e;
  int i, j, n;
  size_t size;
  int update;
  int fd;

  fd = open_tty(port, baud);
  if (fd < 0) { warn("%s", port); return -1; }

  if (mkbl_connect(fd)) return -1;
  dev = mkbl_device(fd);
  if (!dev) return -1;
  data = mkbl_eeprom(fd, dev);
  if (!data) return -1;

  /* magic string */
  if (strcmp(data, TK3_MAGIC)) {
    warnx("invalid eeprom data, please flash with option -i");
    return -1;
  }

  /* read old settings */
  p = data + sizeof(TK3_MAGIC);
  for(n = 0; *p; n++) {
    snprintf(settings[n].key, sizeof(settings[n].key), "%s", p);
    p += strlen(p) + 1;
    settings[n].value = *(unsigned char *)p++<<24;
    settings[n].value|= *(unsigned char *)p++<<16;
    settings[n].value|= *(unsigned char *)p++<<8;
    settings[n].value|= *(unsigned char *)p++;
    snprintf(settings[n].descr, sizeof(settings[n].descr), "%s", p);
    p += strlen(p) + 1;

    if (n > 127) {
      warnx("too many settings");
      return -1;
    }
  }
  free(data);

  /* merge old and new settings */
  update = 0;
  for(i = 0; argv[i] && argv[i+1]; i+=2) {
    if (!*argv[i]) continue;

    for(j = 0; j < n; j++) if (!strcmp(settings[j].key, argv[i])) break;
    if (j >= n) {
      warnx("invalid setting: %s", argv[i]);
      return -1;
    }

    update = 1;
    settings[j].value = strtol(argv[i+1], &e, 0);
    if (*e) {
      warnx("invalid value for key %s: %s", argv[i], argv[i+1]);
      return -1;
    }
  }

  /* re-flash */
  size = sizeof(TK3_MAGIC);
  for(i = 0; i < n; i++) {
    size += strlen(settings[i].key) + 1;
    size += 4;
    size += strlen(settings[i].descr) + 1;
  }
  size++;

  data = malloc(size);
  if (!data) return -1;
  p = data;
  strcpy(p, TK3_MAGIC);
  p += sizeof(TK3_MAGIC);
  for(i = 0; i < n; i++) {
    strcpy(p, settings[i].key);
    p += strlen(p) + 1;
    *p++ = (settings[i].value >> 24) & 0xff;
    *p++ = (settings[i].value >> 16) & 0xff;
    *p++ = (settings[i].value >> 8) & 0xff;
    *p++ = settings[i].value & 0xff;
    strcpy(p, settings[i].descr);
    p += strlen(p) + 1;
  }
  *p = 0;

  print_settings(data + sizeof(TK3_MAGIC));

  if (update) {
    if (mkbl_write(fd, dev, data, size, FL_EEPROM))
      return -1;
  }

  free(data);
  if (mkbl_exit(fd)) return -1;
  return 0;
}


/* --- mkbl_connect -------------------------------------------------------- */

/* connect to the bootloader */

static int
mkbl_connect(int fd)
{
  char data[4];
  int timeout;
  ssize_t l;

  while (read_serial(fd, data, 4, 10) > 0);

  printf("Please power on or reset the device...");
  fflush(stdout);
  timeout = 30;
  do {
    write_serial(fd, "\x1b", 1);
    while (read_serial(fd, data, 4, 10) > 0);
    usleep(20000);

    write_serial(fd, "\xaa", 1);

    l = read_serial(fd, data, 4, 480);
    if (l == 4) break;

    printf("."); fflush(stdout);
  } while(--timeout);
  printf("\n");
  if (l < 4) {
    warnx("timeout while probing the device");
    return -1;
  }

  if (data[0] != 'M' || data[1] != 'K' || data[2] != 'B' || data[3] != 'L') {
    warnx("not a mikrokopter device");
    return -1;
  }

  write_serial(fd, "V", 1);
  l = read_serial(fd, data, 3, 1000);
  if (l != 3) {
    warnx("not a mikrokopter device");
    return -1;
  }

  printf("Bootloader version %c.%c.%c\n", data[0], data[1], data[2]);
  return 0;
}


/* --- mkbl_device --------------------------------------------------------- */

/* check microcontroller device */

static const struct avr_dev *
mkbl_device(int fd)
{
  const struct avr_dev *dev;
  unsigned char c[3], r[2];
  ssize_t l;

  write_serial(fd, "t", 1);
  l = read_serial(fd, &c[1], 2, 1000);
  if (l != 2) { warnx("timeout while checking the device"); return NULL; }

  for(dev = avr_devs; dev->id ; dev++)
    if (dev->id == c[1]) break;
  if (dev->id != c[1]) {
    warnx("unknown device %x", c[1]); return NULL;
  }

  r[0] = r[2] ? c[2] + '@' : '\0';
  r[1] = '\0';
  printf("Probed an %s%s microcontroller\n", dev->name, r);

  c[0] = 'T';
  write_serial(fd, c, 2);
  l = read_serial(fd, c, 1, 1000);
  if (l != 1) { warnx("timeout while checking the device"); return NULL; }
  if (c[0] != 0xd) { warnx("cannot set microcontroller device"); return NULL; }

  return dev;
}


/* --- mkbl_write ---------------------------------------------------------- */

/* flash program */

static int
mkbl_write(int fd, const struct avr_dev *dev, char *data, size_t size,
           enum memk kind)
{
  size_t block, percent;
  unsigned char c[4];
  ssize_t p, l;
  uint16_t crc;

  if (size > ((kind == FL_FLASH) ? dev->flash_size : dev->eeprom_size)) {
    warnx("program does not fit into %s memory",
          (kind == FL_FLASH) ? "flash" : "eeprom");
    return -1;
  }

  /* get write size */
  write_serial(fd, "b", 1);
  l = read_serial(fd, c, 3, 1000);
  if (l != 3 || c[0] != 'Y') {
    warnx("cannot get write buffer size");
    return -1;
  }

  block = (c[1]<<8) + c[2];
  printf("Using %zu bytes writes for a total of %zu bytes\n", block, size);

  if (kind == FL_FLASH) {
    /* clear flash */
    printf("Clearing flash...");
    fflush(stdout);
    write_serial(fd, "e", 1);
    l = read_serial(fd, c, 1, 30000 /* 30s timeout */);
    if (l != 1) { warnx("timeout while clearing flash"); return -1; }
    if (c[0] != 0xd) { warnx("cannot clear flash"); return -1; }
    printf(" ok\n");
  }

  /* write data */
  printf("Writing data...");
  fflush(stdout);
  percent = 0;

  c[0] = 'A';
  c[1] = 0;
  c[2] = 0;
  write_serial(fd, c, 3);
  l = read_serial(fd, c, 1, 1000);
  if (l != 1) { warnx("timeout while setting write address"); return -1; }
  if (c[0] != 0xd) { warnx("cannot set write address"); return -1; }

  for(p = 0; p < size; p += block) {
    int retries = 0;

again:
    l = (p + block > size) ? size-p : block;
    percent += l;
    printf("\rWriting data... %d%%", (int)((double)percent/size*100));
    fflush(stdout);

    c[0] = 'B';
    c[1] = (l >> 8) & 0xff;
    c[2] = l & 0xff;
    c[3] = (kind == FL_FLASH) ? 'F' : 'E';
    write_serial(fd, c, 4);
    write_serial(fd, data + p, l);

    /* writing to flash mem requires checksum */
    if (kind == FL_FLASH) {
      crc = mkbl_crc16(data + p, l);
      c[0] = crc >> 8;
      c[1] = crc & 0xff;
      write_serial(fd, c, 2);
    }

    l = read_serial(fd, c, 1, 1000);
    if (l != 1 || c[0] != 0xd) {
      if (++retries < 3) {
        char dummy[4];
        while (read_serial(fd, dummy, 4, 10) > 0);
        printf("\rError writing data at %d%% (retrying)\n",
               (int)((double)percent/size*100));

        c[0] = 'A';
        c[1] = (p >> 8) & 0xff;
        c[2] = p & 0xff;
        write_serial(fd, c, 3);
        l = read_serial(fd, c, 1, 1000);
        if (l != 1) { warnx("timeout while setting write address"); return -1; }
        if (c[0] != 0xd) { warnx("cannot set write address"); return -1; }

        goto again;
      }
    }

    if (l != 1) { warnx("timeout writing to flash"); return -1; }
    if (c[0] != 0xd) { warnx("cannot write to flash"); return -1; }
  }
  printf("\n");

  return 0;
}


/* --- mkbl_eeprom --------------------------------------------------------- */

/* read data from eeprom */

static void *
mkbl_eeprom(int fd, const struct avr_dev *dev)
{
  unsigned char c[4];
  char *data;
  ssize_t p, l, n, block, percent;

  data = malloc(dev->eeprom_size);
  if (!data) { warnx("out of memory"); return NULL; }

  /* get read size */
  write_serial(fd, "b", 1);
  l = read_serial(fd, c, 3, 1000);
  if (l != 3 || c[0] != 'Y') {
    warnx("cannot get read buffer size");
    return NULL;
  }

  block = (c[1]<<8) + c[2];
  printf("Using %zu bytes reads for a total of %zu bytes\n",
         block, dev->eeprom_size);

  /* read eeprom */
  printf("Reading eeprom data...");
  fflush(stdout);
  percent = 0;

  c[0] = 'A';
  c[1] = 0;
  c[2] = 0;
  write_serial(fd, c, 3);
  l = read_serial(fd, c, 1, 1000);
  if (l != 1) { warnx("timeout while setting read address"); return NULL; }
  if (c[0] != 0xd) { warnx("cannot set read address"); return NULL; }

  for(p = 0; p < dev->eeprom_size; p += block) {
    l = (p + block > dev->eeprom_size) ? dev->eeprom_size-p : block;
    percent += l;
    printf("\rReading eeprom data... %d%%",
           (int)((double)percent/dev->eeprom_size*100));
    fflush(stdout);

    c[0] = 'g';
    c[1] = (l >> 8) & 0xff;
    c[2] = l & 0xff;
    c[3] = 'E';
    write_serial(fd, c, 4);

    n = read_serial(fd, data + p, l, 30000/* 30s timeout */);
    if (n != l) { warnx("timeout reading eeprom"); return NULL; }
  }
  printf("\n");

  return data;
}


/* --- mkbl_exit ----------------------------------------------------------- */

static int
mkbl_exit(int fd)
{
  write_serial(fd, "E", 1);
  return 0;
}
